from nose.tools import assert_equal

from blackbelt.dependencies import (
    parse_dep,
    parse_license_text,
)


class TestDependencyNameParsing(object):

    def test_no_version(self):
        assert_equal(parse_dep('react'), ('react', 'latest'))

    def test_at_separator(self):
        assert_equal(parse_dep('react@16.2'), ('react', '16.2'))

    def test_equals_separator(self):
        assert_equal(parse_dep('react==16.2'), ('react', '16.2'))


class TestLicenseTextParsing(object):

    def test_basic(self):
        assert_equal(parse_license_text('''
            Copyright 2009-2014 Contributors. All rights reserved.

            Permission is hereby granted,
            free of charge...

            ...paragraph
        '''), (
            'Copyright 2009-2014 Contributors. All rights reserved.',
            'Permission is hereby granted, free of charge...\n\n...paragraph',
        ))

    def test_heading(self):
        assert_equal(parse_license_text('''
            MIT License

            Copyright (c) 2013-present, Facebook, Inc.

            Permission is hereby granted,
            free of charge...

            ...paragraph
        '''), (
            'Copyright (c) 2013-present, Facebook, Inc.',
            'Permission is hereby granted, free of charge...\n\n...paragraph',
        ))

    def test_rubbish(self):
        assert_equal(parse_license_text('''
            # Black Belt

            Black belt is collection of scripts, tools and guidelines used for developing projects The Apiary Way.

            ## Installation & Usage

            Please refer to [The Black Belt Documentation](http://black-belt.readthedocs.org/).
        '''), (
            None,
            None,
        ))

    def test_empty(self):
        assert_equal(parse_license_text(''), (
            None,
            None,
        ))

    def test_none(self):
        assert_equal(parse_license_text(None), (
            None,
            None,
        ))



import json
from os.path import expanduser


try:
    configFile = expanduser('~/.blackbelt')
    with open(configFile) as f:
        config = json.loads(f.read())
except IOError:
    config = {
        'trello': {
            'access_token': None,
            'work_board_url': None,
            'work_board_id': None,
            'product_board_url': None,
            'product_board_id': None,
            'tea_board_url': None,
            'tea_board_id': None
        },
        'github': {
            'access_token': None
        },
        'circleci': {
            'access_token': None
        },
        'slack': {
            'access_token': None
        },
        'office': {
            'office_manager_name': None,
            'office_manager_email': None
        }
    }

# default config, overwrite/modularize
if not 'work_board_id' in config['trello'] or not config['trello']['work_board_id']:
    config['trello']['work_board_id'] = '1KsoiV9e'

# default config, overwrite/modularize
if not 'tea_board_id' in config['trello'] or not config['trello']['tea_board_id']:
    config['trello']['tea_board_id'] = '3ShcntV4'

if not 'office' in config:
    config['office'] = {}

if not 'office_manager_name' in config['office'] or not config['office']['office_manager_name']:
    config['office']['office_manager_name'] = 'Alzbeta'

if not 'office_manager_email' in config['office'] or not config['office']['office_manager_email']:
    config['office']['office_manager_email'] = 'alzbeta@apiary.io'

config['trello']['work_column_name'] = 'Doing'
config['trello']['pause_column_name'] = 'Ready'



from mock import patch

from blackbelt.messages import post_message

@patch("blackbelt.messages.slack_post_message")
def test_post_message(mock_post_message):
  post_message('test', 'room')
  assert mock_post_message.called, True
  assert mock_post_message.call_args_list, ['test', 'room']



import os

import click

from blackbelt.dependencies import check as do_check, run, parse_dep


@click.group(help='Dependency management')
def cli():
    pass


def validate_dep(ctx, param, dep):
    try:
        dep_name, dep_version = parse_dep(dep)
    except Exception:
        raise click.BadParameter('The dependency format should be e.g. react@16.2')
    else:
        if dep_version == 'latest':
            dep_version = run(['npm', 'view', dep_name, 'version'])
        return (dep_name, dep_version)


@cli.command()
@click.argument('dep', callback=validate_dep)
@click.option('--dev/--no-dev', default=False,
              help='Whether to include dev dependencies.')
@click.option('--debug/--no-debug', default=False,
              help='Print debug information.')
@click.option('--list-path', type=click.File(mode='w'),
              default=lambda: os.path.join(os.getcwd(), 'list.txt'),
              help='Where to save the list of 4th party deps.')
@click.option('--licenses-path', type=click.File(mode='w'),
              default=lambda: os.path.join(os.getcwd(), 'licenses.txt'),
              help='Where to save the Public License field contents.')
def check(*args, **kwargs):
    do_check(*args, **kwargs)



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

import re
import os
import json
import shutil
import tempfile
import subprocess
import itertools
import operator

import click


__all__ = ['check', 'parse_dep']


# https://github.com/davglass/license-checker#custom-format
LICENSE_CHECKER_FORMAT_KEYS = [
    'name', 'repository', 'dependencyPath', 'url', 'version', 'licenses',
    'licenseText', 'licenseModified', 'licenseFile', 'path',
]

# https://github.com/davglass/license-checker#custom-format
LICENSE_CHECKER_FORMAT = {key: None for key in LICENSE_CHECKER_FORMAT_KEYS}

# Search Oracle wiki for 'Licenses Eligible for Pre-Approval - Distribution'
# List of official license codes: https://spdx.org/licenses/
LICENSES_DISTRIBUTED = [
    'Apache', 'Apache-1.1', 'Apache-2.0', '0BSD', 'BSD', 'BSD-2-Clause',
    'BSD-3-Clause', 'ISC', 'MIT', 'PHP-3.0', 'UPL', 'UPL-1.0', 'ZPL-2.0',
    'Unlicense', 'Python-2.0', 'FreeBSD', 'Apache License, Version 2.0',
]
LICENSES_DISTRIBUTED += [license + '*' for license in LICENSES_DISTRIBUTED]

MISSING_COPYRIGHT_NOTICE_WARNING = '!!! MISSING COPYRIGHT NOTICE !!!'
MISSING_LICENSE_TEXT_WARNING = '!!! MISSING LICENSE !!!'


def check(dep, list_path, licenses_path, dev=False, debug=False):
    """
    Checks the dependency for licenses and vulnerabilities before you can
    add it to the Third-Party Approval Process:

    #. Installs the dependency in a temporary directory
    #. Checks licenses
    #. Checks for ``PATENTS`` files (TODO)
    #. Provides you with:

        - The list of 4th party deps for the Technology Usage Note field
        - The contents of the Public License field
        - List of packages not eligible for Pre-Approval
        - List of known vulnerabilities (TODO)

    Example::

        bb dep check react@16.2

    Requirements:

    - ``npm install -g license-checker``
    - ``npm install -g licensecheck`` (TODO: ``licensecheck --once --dev --tsv``)
    - ``npm install -g nsp`` (TODO: ``nsp check``)
    - ``npm install -g snyk && snyk auth`` (TODO: ``snyk test``)
    """
    ensure_executables(['npm', 'license-checker'])

    click.echo('Analyzing the package...')
    dep_name, dep_version = dep

    with tempfile.TemporaryDirectory() as tmp_dir:
        install(dep_name, dep_version, tmp_dir, dev=dev)
        licenses = license_checker(tmp_dir)

    pre_approval_verdict = get_pre_approval_verdict(licenses)
    details, fourth_party_licenses = separate_top_level_details(licenses, dep_name)

    click.echo('Creating the list of 4th party deps...')
    list_path.write(create_deps_list(fourth_party_licenses))
    click.echo('Creating the Public License field contents...')
    licenses_path.write(create_licenses_list(details, fourth_party_licenses))

    color = 'green' if pre_approval_verdict else 'red'
    click.secho('\n{name}@{version}'.format(**details), bold=True, fg=color)
    click.echo((
        'License: {licenses}\n'
        'Dependencies: {dependencies}\n'
        'Elligible for Pre-Approval: {pre_approval_verdict}'
    ).format(
        licenses=details['licenses'],
        dependencies=len(fourth_party_licenses),
        pre_approval_verdict=pre_approval_verdict,
    ))

    problematic_licenses = [
        details for details in licenses
        if details['not_pre_approved_reasons']
    ]
    if problematic_licenses:
        heading = '\nProblematic Licenses: {0}'.format(len(problematic_licenses))
        click.secho(heading, bold=True, fg=color)
        missing = False

        for details in problematic_licenses:
            reasons = ', '.join(details['not_pre_approved_reasons'])
            missing = missing or 'missing' in reasons

            line = click.style('{name} {version} ({licenses})'.format(**details), bold=True)
            click.echo('{0} - {1}'.format(line, reasons))

            if debug:
                click.echo(' ・ npm: https://www.npmjs.com/package/{0}'.format(details['name']))
                if details.get('repo'):
                    click.echo(' ・ repo: {0}'.format(details['repo']))
                if details.get('license_file'):
                    click.echo(' ・ license file: {0}'.format(details['license_file']))

        if missing:
            click.echo(
                '\nBad luck! Before adding the dependency to the approval '
                'process you need to manually go through the dependencies, '
                'get the missing info and complete the generated files '
                'with it.'
            )
        if not debug:
            click.echo('\nProTip: You can use --debug to print more details.')


def install(dep_name, dep_version, project_dir, dev=False):
    click.echo('Getting dependencies...')
    dep = '{0}@{1}'.format(dep_name, dep_version)
    run(['npm', 'install', dep], cwd=project_dir)
    if dev:
        click.echo('Getting dev dependencies...')
        shutil.copy(
            os.path.join(project_dir, 'node_modules', dep_name, 'package.json'),
            os.path.join(project_dir),
        )

        package_json = os.path.join(project_dir, 'package.json')
        with open(package_json) as f:
            package_data = json.load(f)

        package_data_just_deps = {}
        for key, value in package_data.items():
            if 'dependencies' in key.lower():
                package_data_just_deps[key] = value
        with open(package_json, 'w') as f:
            json.dump(package_data_just_deps, f)

        run(['npm', 'install'], cwd=project_dir)


def get_pre_approval_verdict(licenses):
    return all(not details['not_pre_approved_reasons'] for details in licenses)


def create_deps_list(licenses):
    return '\n'.join([
        '{name} {version} ({licenses})'.format(**details)
        for details in licenses
    ])


# See the internal FAQ and search for following questions:
#
# - What information is required in the Public License field?
# - What's the best way to format all the information in the Public
#   License field?
# - If there are multiple dependencies licensed under the same terms,
#   do I need to repeat those terms for every dependency?
def create_licenses_list(top_level_details, fourth_party_licenses):
    sections = []

    sections.append(top_level_details['copyright_notice'])
    sections.append(top_level_details['license_text'])

    key_fn = operator.attrgetter('license_text')
    licenses = sorted(fourth_party_licenses, key=key_fn)
    identical_license_groups = itertools.groupby(licenses, key_fn)

    for license_text, details_list in identical_license_groups:
        details_list = list(details_list)
        if len(details_list) == 1:
            section = (
                '{name} {version} ({licenses})\n'
                '{copyright_notice}\n'
                '{license_text}'
            ).format(**details_list[0])
            sections.append(section)
        else:
            for details in details_list:
                section = (
                    '{name} {version} ({licenses})\n'
                    '{copyright_notice}'
                ).format(**details)
                sections.append(section)

            section = ''.join([
                '{name} {version}\n'.format(**details)
                for details in details_list
            ]) + '\n' + license_text
            sections.append(section)

    separator = '\n---------separator---------\n'
    return separator.join(sections) + separator


def separate_top_level_details(licenses, dep_name):
    dep_details = None
    fourth_party_licenses = []

    for details in licenses:
        if details['name'] == dep_name:
            dep_details = details
        else:
            fourth_party_licenses.append(details)

    return (dep_details, fourth_party_licenses)


def license_checker(project_dir):
    format_file = os.path.join(project_dir, 'format.json')
    with open(format_file, 'w') as f:
        json.dump(LICENSE_CHECKER_FORMAT, f)

    args = ['--unknown', '--json', '--customPath', format_file]
    output = run(['license-checker'] + args, cwd=project_dir)

    licenses = []
    for details in json.loads(output).values():
        copyright_notice, license_text = parse_license_text(details.get('licenseText'))
        license_names = parse_license_names(details.get('licenses'))

        details = {
            'name': details['name'],
            'version': details['version'],
            'repo': details.get('repository'),
            'license_file': parse_license_filename(details['name'], details.get('licenseFile')),
            'copyright_notice': copyright_notice or MISSING_COPYRIGHT_NOTICE_WARNING,
            'license_text': license_text or MISSING_LICENSE_TEXT_WARNING,
            'licenses': license_names,
        }
        details['not_pre_approved_reasons'] = check_pre_approval_elligibility(details)

        licenses.append(details)
    return licenses


def ensure_executables(executables):
    for executable in executables:
        try:
            # check=False, because some programs return non-zero status when
            # they print --version output
            run([executable, '--version'], check=False)
        except FileNotFoundError:
            if executable == 'npm':
                msg = "'npm' is needed, but it's not installed."
                click.echo(msg, err=True)
                raise click.Abort()
            else:
                confirm = (
                    "'{0}' is needed, but it's not installed. "
                    "Install by 'npm install -g {0}'?"
                )
                click.confirm(confirm.format(executable), abort=True)
                run(['npm', 'install', '-g', executable])


# Unfortunately, we cannot use subprocess.run(), because BB still supports Py2
def run(args, cwd=None, check=True):
    kwargs = {'cwd': cwd}
    try:
        with open(os.devnull, 'w') as devnull:
            kwargs['stderr'] = devnull
            output = subprocess.check_output(args, **kwargs)
    except subprocess.CalledProcessError:
        if check:
            raise
    else:
        return output.decode().strip()


def parse_license_names(value):
    try:
        return value.lstrip('(').rstrip(')')
    except AttributeError:
        # Someone is using the deprecated 'licenses' field, we got a list.
        # Since we have no idea whether these licenses should be written
        # with OR, AND, ... let's join it by commas.
        #
        # See also https://docs.npmjs.com/files/package.json#license
        return ', '.join(filter(None, value))


def parse_license_text(text):
    copyright_notice = None
    license_text = None

    for line in (text or '').splitlines():
        if re.search(r'copyright|\(c\)|©', line, re.I):
            copyright_notice = line.strip()
            break

    if copyright_notice:
        license_text = text.split(copyright_notice)[1]
        license_text = license_text.strip()
        license_text = re.sub(r' +', ' ', license_text)
        license_text = re.sub(r' ?(\r\n|\n)+ ?', remove_newlines_keep_paragraps, license_text)

    return (copyright_notice, license_text)


def remove_newlines_keep_paragraps(match):
    newlines = re.findall(r'\r\n|\n', match.group(0))
    if len(newlines) > 1:
        return ''.join(newlines[:2])
    return ' '


def parse_license_filename(dep_name, path):
    if path:
        try:
            return path.split('node_modules')[1].lstrip('/')
        except IndexError:
            return path
    return None


def check_pre_approval_elligibility(details):
    reasons = []
    if not details['licenses'] in LICENSES_DISTRIBUTED:
        reasons.append('license not pre-approved')
    if details['copyright_notice'] == MISSING_COPYRIGHT_NOTICE_WARNING:
        reasons.append('missing copyright notice')
    if details['license_text'] == MISSING_LICENSE_TEXT_WARNING:
        reasons.append('missing full license text')
    return reasons


def parse_dep(dep):
    split_result = re.split(r'==|@', dep)
    try:
        return (split_result[0], split_result[1])
    except IndexError:
        return (split_result[0], 'latest')



from nose.tools import assert_equal, assert_raises
from mock import patch, MagicMock
import requests
import json

from blackbelt.handle_github import (
    get_pr_info,
    verify_merge,
    get_pr_ticket_id,
    verify_pr_state,
    verify_pr_required_checks,
    verify_branch_required_checks,
)


class TestGithubPullRequestParsing(object):

    github_pr = 'https://github.com/apiaryio/apiary-test/pull/123'

    def setUp(self):
        self.parsed = get_pr_info(self.github_pr)

    def test_repo_owner(self):
        assert_equal('apiaryio', self.parsed['owner'])

    def test_repo_name(self):
        assert_equal('apiary-test', self.parsed['name'])

    def test_pr_number(self):
        assert_equal('123', self.parsed['number'])

    def test_files_link(self):
        parsed = get_pr_info(self.github_pr + '/files')
        assert_equal('123', parsed['number'])


    def test_trello_id_extracted(self):
        link = get_pr_ticket_id("""
        # This is edited and a long pull requests
        Hoever, I'd like to warn you...

        Pull request for [Naming fixes](https://trello.com/c/yx2SNE3J/1910-naming-fixes).
        """)

        assert_equal('yx2SNE3J', link)

    def test_error_on_bad_phrase(self):
        assert_raises(ValueError, lambda: get_pr_ticket_id("""
        Related to [naming fixes](https://trello.com/c/yx2SNE3J/1910-naming-fixes)
        """))

class TestMergeVerificationRetries(object):
    pr_url = 'https://github.com/apiaryio/apiary-test/pull/123'

    headers = {}

    @patch.object(requests, 'get')
    def test_timeout(self, get_method):
        get_method.return_value = fake_response = MagicMock()
        fake_response.status_code = 404

        assert_raises(ValueError, lambda: verify_merge(
            self.pr_url,
            self.headers,
            max_waiting_time=0.01,
            retry_time=0.001
        ))

class TestStatusVerifications(object):
    pr_url = 'https://github.com/apiaryio/apiary-test/pull/123'
    branch_name = 'user/local-branch'

    def setUp(self):
        with open('./test/gh-mock-response/pr-details-success.json') as json_data:
            self.pr_details_open = json.load(json_data)
        with open('./test/gh-mock-response/pr-details-failure.json') as json_data:
            self.pr_details_closed = json.load(json_data)

    def init_mock(self, json_data_file):
        fake_response = MagicMock()
        fake_response.status_code = 200
        with open(json_data_file) as json_data:
            response_data = json.load(json_data)
            fake_response.json.return_value = response_data
        return fake_response

    @patch.object(requests, 'get')
    def test_verify_branch_required_checks_success(self, get_method):
        get_method.return_value = self.init_mock(json_data_file='./test/gh-mock-response/ref-status-success.json')
        message = verify_branch_required_checks(self.branch_name)
        assert_equal('Branch required checks (2): success', message)

    @patch.object(requests, 'get')
    def test_verify_branch_required_checks_failure(self, get_method):
        get_method.return_value = self.init_mock(json_data_file='./test/gh-mock-response/ref-status-failure.json')
        assert_raises(ValueError, lambda: verify_branch_required_checks(self.branch_name))

    @patch.object(requests, 'get')
    def test_verify_pr_status_success(self, get_method):
        get_method.return_value = self.init_mock(json_data_file='./test/gh-mock-response/pr-details-success.json')
        message = verify_pr_state(self.pr_details_open)
        assert_equal('PR state: open', message)

    @patch.object(requests, 'get')
    def test_verify_pr_status_failure(self, get_method):
        get_method.return_value = self.init_mock(json_data_file='./test/gh-mock-response/pr-details-failure.json')
        assert_raises(ValueError, lambda: verify_pr_state(self.pr_details_closed))

    @patch.object(requests, 'get')
    def test_verify_pr_required_checks_success(self, get_method):
        get_method.return_value = self.init_mock(json_data_file='./test/gh-mock-response/ref-status-success.json')
        message = verify_pr_required_checks(self.pr_details_open)
        assert_equal('PR required checks (2): success', message)

    @patch.object(requests, 'get')
    def test_verify_pr_required_checks_failure(self, get_method):
        get_method.return_value = self.init_mock(json_data_file='./test/gh-mock-response/ref-status-failure.json')
        assert_raises(ValueError, lambda: verify_pr_required_checks(self.pr_details_closed))


class TestGitHubPullRequestResponseData(object):

    def test_github_details_response_data(self):
        with open('./test/gh-mock-response/pr-details-success.json') as json_data:
            pr_details = json.load(json_data)

        assert_equal('octocat/Hello-World', pr_details['base']['repo']['full_name'])
        assert_equal('new-topic', pr_details['head']['ref'])
        assert_equal('git:github.com/octocat/Hello-World.git', pr_details['base']['repo']['git_url'])
        assert_equal('git@github.com:octocat/Hello-World.git', pr_details['base']['repo']['ssh_url'])
        assert_equal('https://github.com/octocat/Hello-World.git', pr_details['base']['repo']['clone_url'])
        assert_equal('6dcb09b5b57875f334f61aebed695e2e4193db5e', pr_details['head']['sha'])
        assert_equal('new-topic', pr_details['head']['ref'])
        assert_equal(1347, pr_details['number'])
        assert_equal('new-feature', pr_details['title'])
        assert_equal(10, pr_details['comments'])
        assert_equal(3, pr_details['commits'])
        assert_equal('octocat', pr_details['base']['repo']['owner']['login'])
        assert_equal('Hello-World', pr_details['base']['repo']['name'])
        assert_equal('Please pull these awesome changes', pr_details['body'])
        assert_equal('https://github.com/octocat/Hello-World/pull/1347', pr_details['html_url'])
        assert_equal('open', pr_details['state'])



import click
import os

plugin_folder = os.path.join(os.path.dirname(__file__), 'commands')

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

class BlackBelt(click.MultiCommand):

    def list_commands(self, ctx):
        rv = []
        for filename in os.listdir(plugin_folder):
            if filename.endswith('.py') and filename != '__init__.py':
                rv.append(filename[:-3])
        rv.sort()
        return rv

    def get_command(self, ctx, name):
        ns = {}
        fn = os.path.join(plugin_folder, name + '.py')
        if os.path.exists(fn):
            with open(fn) as f:
                code = compile(f.read(), fn, 'exec')
                eval(code, ns, ns)
            return ns['cli']

cli = BlackBelt(context_settings=CONTEXT_SETTINGS, help='Black Belt: automate project The Apiary Way. Please provide a command.')


# backward compatibility
def main():
    cli()

if __name__ == '__main__':
    cli()



import click

from blackbelt.deployment import (
    deploy_production
)


@click.group(invoke_without_command=True, help="""
    Deploy master to production
    """)
def cli():
    deploy_production()
